
/*
 * This file is part of fdnpkg.
 * 
 * Loads the list of repositories from the config file specified in %FDNPKG%.
 * Returns the amount of repositories found (and loaded) on success, or -1 on failure.
 * 
 * Copyright (C) Mateusz Viste 2012
 */

#include <stdio.h>  /* printf(), fclose(), fopen()... */
#include <string.h> /* strcasecmp() */
#include <stdlib.h> /* malloc(), free() */


void freeconf(char **repolist, int repscount) {
  int x;
  for (x = 0; x < repscount; x++) free(repolist[x]);
}


static int checkfordoubledrepos(char **repolist, int repocount) {
  int x, y;
  for (x = 0; x < (repocount - 1); x++) {
    for (y = x + 1; y < repocount; y++) {
      if (strcmp(repolist[x], repolist[y]) == 0) {
        printf("Error: repository '%s' is listed twice!\r\n", repolist[x]);
        return(-1);
      }
    }
  }
  return(0);
}


int loadconf(char *cfgfile, char **repolist, int maxreps) {
  int bytebuff, parserstate = 0;
  FILE *fd;
  #define maxtok 16
  char token[maxtok];
  #define maxval 1024
  char value[maxval];
  int curtok = 0, curval = 0, nline = 1;
  int repocount = 0;
  fd = fopen(cfgfile, "r");
  if (fd == NULL) {
    printf("Could not open config file '%s'!\n", cfgfile);
    return(-1);
  }

  /* read the config file line by line */
  do {
    bytebuff = fgetc(fd);
    if (bytebuff != '\r') {
      switch (parserstate) {
        case 0: /* Looking for start of line */
          if ((bytebuff == EOF) || (bytebuff == ' ') || (bytebuff == '\t')) break;
          if (bytebuff == '\n') {
            nline += 1;
            break;
          }
          if (bytebuff == '#') {
              parserstate = 9;
            } else {
              token[0] = bytebuff;
              curtok = 1;
              parserstate = 1;
          }
          break;
        case 1: /* Looking for token end */
          if ((bytebuff == EOF) || (bytebuff == '\n')) {
            printf("Warning: token without value on line #%d\n", nline);
            if (bytebuff == '\n') nline += 1;
            parserstate = 0;
            break;
          }
          if ((bytebuff == ' ') || (bytebuff == '\t')) {
              token[curtok] = 0;
              parserstate = 2;
            } else {
              token[curtok] = bytebuff;
              curtok += 1;
              if (curtok >= maxtok) {
                parserstate = 9; /* ignore the line */
                printf("Warning: Config file token overflow on line #%d\n", nline);
              }
          }
          break;
        case 2: /* Looking for value start */
          if ((bytebuff == EOF) || (bytebuff == '\n')) {
            printf("Warning: token with empty value on line #%d\n", nline);
            if (bytebuff == '\n') nline += 1;
            parserstate = 0;
            break;
          }
          if ((bytebuff != ' ') && (bytebuff != '\t')) {
            value[0] = bytebuff;
            curval = 1;
            parserstate = 3;
          }
          break;
        case 3: /* Looking for value end */
          if ((bytebuff == EOF) || (bytebuff == '\n')) {
              parserstate = 0;
              value[curval] = 0;
              if ((value[curval - 1] == ' ') || (value[curval - 1] == '\t')) {
                printf("Warning: trailing white-space(s) after value on line #%d\n", nline);
                while ((value[curval - 1] == ' ') || (value[curval - 1] == '\t')) value[--curval] = 0;
              }
              /* Interpret the token/value pair now! */
              /* printf("token='%s' ; value = '%s'\n", token, value); */
              if (strcasecmp(token, "REPO") == 0) { /* Repository declaration */
                  if (repocount >= maxreps) {
                      printf("Dropped a repository: too many configured (max=%d)\n", maxreps);
                    } else {
                      repolist[repocount] = malloc(curval + 1); /* +1 for a trailing / if not specified */
                      if (repolist[repocount] == NULL) {
                        puts("Out of memory while reading repositories list");
                        freeconf(repolist, repocount);
                        return(-1);
                      }
                      if (value[curval - 1] != '/') { /* add a trailing slash to the url if not there already */
                          sprintf(repolist[repocount], "%s/", value);
                        } else { 
                          sprintf(repolist[repocount], "%s", value);
                      }
                      repocount += 1;
                  }
                } else { /* unknown token */
                  printf("Warning: Unknown token '%s' at line #%d\n", token, nline);
              }
              /* interpretation done */
              if (bytebuff == '\n') nline += 1;
            } else {
              value[curval] = bytebuff;
              curval += 1;
              if ((curval + 1) >= maxval) {
                parserstate = 9; /* ignore the line */
                printf("Warning: Config file value overflow on line #%d\n", nline);
              }
          }
          break;
        case 9: /* Got comment, ignoring the rest of line */
          if (bytebuff == EOF) break;
          if (bytebuff == '\n') {
            nline += 1;
            parserstate = 0;
          }
          break;
      }
    }
  } while (bytebuff != EOF);
  fclose(fd);

  /* Look out for doubled repositories */
  if (checkfordoubledrepos(repolist, repocount) != 0) return(-1);

  return(repocount);
}
