/*
 * Package database manipulation routines.
 * This file is part of fdnpkg
 * 
 * Copyright (C) Mateusz Viste 2012
 */


#include <stdio.h>  /* printf(), puts()... */
#include <stdlib.h> /* malloc(), free()... */
#include <string.h> /* strcmp() */
#include "getdelim.h" /* a getdelim() implementation - there is none availble on DJGPP */
#include "pkgdb.h"


/* Free the memory occupied by the linked list */
void freedb(struct pkgdb **db) {
  struct pkgdb *nextdb, *curdb;
  struct pkgrepo *nextrepo, *currepo;
  curdb = *db;
  while (curdb != NULL) {
    nextdb = curdb->nextpkg;
    /* free all possible repositories */
    currepo = curdb->repolist;
    while (currepo != NULL) {
      nextrepo = currepo->nextrepo;
      free(currepo);
      currepo = nextrepo;
    }
    free(curdb);
    curdb = nextdb;
  }
}


/* Create the initial stucture for the db (first node) */
struct pkgdb *createdb() {
  struct pkgdb *newentry;
  newentry = malloc(sizeof(struct pkgdb));
  if (newentry == NULL) return(NULL);
  newentry->repolist = NULL;
  newentry->nextpkg = NULL;
  return(newentry);
}


/* adds a new entry to the db - returns a pointer to the new entry on success, NULL otherwise */
static struct pkgdb *addpkg(struct pkgdb *db, char *pkgname, char *desc) {
  struct pkgdb *oldnextdb, *newentry;
  newentry = createdb();
  if (newentry == NULL) return(NULL);
  strncpy(newentry->name, pkgname, 8);
  newentry->name[8] = 0; /* strncpy doesn't terminate strings if limit reached */
  strncpy(newentry->desc, desc, 63);
  newentry->desc[63] = 0; /* strncpy doesn't terminate strings if limit reached */
  if (strlen(desc) > 63) { /* if desc was longer than 63 bytes, terminate it nicely */
    newentry->desc[60] = '.';
    newentry->desc[61] = '.';
    newentry->desc[62] = '.';
  }
  oldnextdb = db->nextpkg;
  db->nextpkg = newentry;
  newentry->nextpkg = oldnextdb;
  return(newentry);
}


/* adds a new repo to entry *db - returns the new entry address on success, NULL otherwise */
static struct pkgrepo *addrepo(struct pkgdb *pkg, char *repo, char *version) {
  struct pkgrepo *oldnextrepo, *newrepo;
  newrepo = malloc(sizeof(struct pkgrepo));
  if (newrepo == NULL) return(NULL);
  newrepo->nextrepo = pkg->repolist;
  strncpy(newrepo->version, version, 15);
  newrepo->version[15] = 0; /* strncpy doesn't terminate strings if limit reached */
  newrepo->repo = repo;
  oldnextrepo = pkg->repolist;
  pkg->repolist = newrepo;
  newrepo->nextrepo = oldnextrepo;
  return(newrepo);
}


/* Searches after a pkg named 'pkgname' in the db. Returns a pointer to the package if found, or NULL otherwise.
   If no package found, then lastmatch is set to the last alphabetically inferior package */
struct pkgdb *findpkg(struct pkgdb *db, char *pkgname, struct pkgdb **lastmatch) {
  struct pkgdb *curpkg;
  int cmpres;
  *lastmatch = db;
  for (curpkg = db->nextpkg; curpkg != NULL; curpkg = curpkg->nextpkg) {
    cmpres = strcasecmp(curpkg->name, pkgname);
    if (cmpres < 0) {    /* if alphabetically inferior, remember lastmatch */
        *lastmatch = curpkg;
      } else if (cmpres == 0) { /* if equal, then we found what we came for! */
        return(curpkg);
      } else { /* if superior, then there's no point in loosing more time here (assuming the list is sorted) */
        return(NULL);
    }
  }
  return(NULL);
}


static void getpkgmetadata(char *lineptr, char **pkgmeta) {
  int x = 0, y = 1, lastnonemptychar = 0;
  pkgmeta[0] = lineptr;
  pkgmeta[1] = lineptr;
  pkgmeta[2] = lineptr;
  while (lineptr[x] != 0) {
    /* check out what character we have, and remember if it's whitespace or not */
    switch (lineptr[x]) {
      case '\t': /* delimiters */
      case '\r':
      case '\n':
        lineptr[lastnonemptychar + 1] = 0;
        if (y < 3) pkgmeta[y] = &lineptr[x + 1];
        lastnonemptychar = x + 1;
        y++;
        if (y > 3) return;
        break;
      case ' ': /* whitespace... */
        break;
      default: /* remember last non-whitespace char */
        lastnonemptychar = x;
    }
    x++;
  }
}


/* Load the db from file - returns 0 on success, non-zero on failure */
int loaddb(struct pkgdb *db, char *datafile, char *repo) {
  FILE *fd;
  char *lineptr;
  char *pkgmeta[3];
  struct pkgdb *pkg, *lastmatch;
  int result = 0;
  size_t getdelimcount = 0;
  int getdelimlen;
  fd = fopen(datafile, "r");
  if (fd == NULL) return(-1);
  
  /* Check whether the index file has a valid header */
  lineptr = NULL;
  getdelimlen = getdelim(&lineptr, &getdelimcount, '\n', fd);
  if (getdelimlen < 0) {
      free(lineptr);
      puts("Error: invalid index file (bad hdr)! Repository skipped.");
      result = -6;
    } else {
      getpkgmetadata(lineptr, pkgmeta); /* pkgmeta[0] = "FD-REPOv1" */
      if (strcmp(pkgmeta[0], "FD-REPOv1") != 0) {
          free(lineptr);
          puts("Error: invalid index file (bad sig)! Repository skipped.");
          result = -7;
        } else {
          /* start reading the index content */
          for (;;) {
            /* read line from file */
            lineptr = NULL;
            getdelimlen = getdelim(&lineptr, &getdelimcount, '\n', fd);
            if (getdelimlen < 0) {
              free(lineptr);
              break;
            }
            /* printf("Got line: %s\r\n", lineptr); */
            if ((lineptr[0] == 0) || (lineptr[0] == '\r') || (lineptr[0] == '\n')) continue; /* skip empty lines */
            /* get pkgname and version */ /* PkgName <tab> version <tab> description */
            getpkgmetadata(lineptr, pkgmeta); /* pkgmeta[0] = PkgName ; pkgmeta[1] = version ; pkgmeta[2] = description */
            /* look if package already exists in db... */
            pkg = findpkg(db, pkgmeta[0], &lastmatch);
            /* if not -> add the package to the db first */
            if (pkg == NULL) {
              pkg = addpkg(lastmatch, pkgmeta[0], pkgmeta[2]);
              if (pkg == NULL) {
                puts("Out of memory while loading package database!");
                free(lineptr);
                result = -1;
                break;
              }
            }
            /* now add the repository to the package */
            if (addrepo(pkg, repo, pkgmeta[1]) == NULL) { /* ooops, something went wrong */
              puts("Out of memory while loading package database!");
              result = -1;
              free(lineptr);
              break;
            }
            free(lineptr); /* free the memory occupied by the line */
          }
      }
  }
  fclose(fd);
  return(result);
}
